/*
 * tmos - mb-slave-net.c
 *
 * 	Generic modbus slave (network version)
 * 
 * Copyright (C) 2005..2010  Krzysztof Mazur <krzysiek@podlesie.net>
 * Copyright (C) 2006, 2007, 2009  Marek Piwonski <marek@microster.pl>
 * Copyright (C) 2006  Krzysztof Kasperski <casper@microster.pl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/config.h>
#include <tmos/crc16.h>
#include <tmos/modbus.h>
#include <tmos/mb-slave.h>
#include <tmos/net.h>

/*
 * This implementation should be simple. Code size and CPU time
 * is more important than unused features.
 */

/**
 * send_error_message - send ModBUS exception
 * @ex: exception, see MBEX_*
 */
static net_size_t send_error_message(uint8_t __kernel *msg, unsigned char ex)
	__reentrant
{
	msg[1] |= 0x80;
	msg[2] = ex;
	return 3;
}

#ifdef CONFIG_MB_SLAVE_HOLDING_REGISTERS
/**
 * parse_10_msg - parse 0x10 message - write holding registers
 * @msg_len: message length
 */
static net_size_t parse_10_msg(void __kernel *buf, unsigned char msg_len)
	__reentrant
{
	struct modbus_10_cmd __kernel *msg = buf;
	uint8_t i;
	mb_addr_t addr;
	mb_size_t len;
	mb_ex_t ex;
	
	/* ignore messages with incorrect length */
	if (sizeof(*msg) + msg->bytes != msg_len)
		return 0;

	/* -1 to catch also zero length */
	ex = MBEX_INVAL;
	if (msg->len_hi || (msg->len_lo - 1 > 0x7a))
		goto err;
	len = msg->len_lo;

#if CONFIG_MB_SLAVE_SHORT_ADDR
	ex = MBEX_FAULT;
	if (msg->addr_hi)
		goto err;
	addr = msg->addr_lo;
#else
	addr = pack16(msg->addr_lo, msg->addr_hi);
#endif
	
	ex = mb_slave_hr_write(msg->payload, addr, len);
	if (ex != MBEX_OK)
		goto err;

	return sizeof(struct modbus_10_resp);
err:
	return send_error_message(buf, ex);
}

/**
 * parse_6_msg - parse 0x6 message - write holding register
 * @len: message length
 */
static net_size_t parse_06_msg(void __kernel *buf, unsigned char len)
	__reentrant
{
	struct modbus_06 __kernel *msg = buf;
	mb_addr_t addr;
	mb_ex_t ex;

	/* ignore invalid messages */
	if (sizeof(*msg) != len)
		return 0;

#if CONFIG_MB_SLAVE_SHORT_ADDR
	ex = MBEX_FAULT;
	if (msg->addr_hi)
		goto err;
	addr = msg->addr_lo;
#else
	addr = pack16(msg->addr_lo, msg->addr_hi);
#endif

	ex = mb_slave_hr_write(&msg->value, addr, 1);
	if (ex != MBEX_OK)
		goto err;

	return sizeof(struct modbus_06) - 2;
err:
	return send_error_message(buf, ex);
}

#endif

#if CONFIG_MB_SLAVE_HOLDING_REGISTERS || CONFIG_MB_SLAVE_INPUT_REGISTERS
/**
 * parse_03_msg - parse 0x3 message - read holding registers
 * @msg_len: message length
 */
static net_size_t parse_03_msg(void __kernel *buf, unsigned char msg_len,
		unsigned char buf_len)
	__reentrant
{
	struct modbus_03_cmd __kernel *msg = buf;
	struct modbus_03_resp __kernel *resp = buf;
	mb_addr_t addr;
	mb_size_t len;
	mb_ex_t ex;
	unsigned char i;

	/* ignore invalid requests */	
	if (msg_len != sizeof(*msg))
		return 0;

	buf_len -= sizeof(*resp);
	buf_len /= 2;
#if 0
	if (buf_len > 0x7c)
		buf_len = 0x7c;
#endif
	/* -1 to catch also zero length */
	ex = MBEX_INVAL;
	if (msg->len_hi || (msg->len_lo - 1 > buf_len))
		goto err;
	len = msg->len_lo;

#if CONFIG_MB_SLAVE_SHORT_ADDR
	ex = MBEX_FAULT;
	if (msg->addr_hi)
		goto err;
	addr = msg->addr_lo;
#else
	addr = pack16(msg->addr_lo, msg->addr_hi);
#endif

	ex = mb_slave_hr_read(resp->payload, addr, len);
	if (ex != MBEX_OK)
		goto err;
	resp->bytes = len * 2;
	return sizeof(struct modbus_03_resp) + resp->bytes;
err:
	return send_error_message(buf, ex);
}
#endif

/**
 * modbus_xmit - process Modbus network packet
 * FIXME: check for buffer overflow
 */
net_size_t modbus_xmit(struct net_msg __kernel *msg)
	__reentrant
{
	if (msg->tail < 2)
		return 0;

	switch (((uint8_t __kernel *) msg->buffer)[1]) {
#if CONFIG_MB_SLAVE_HOLDING_REGISTERS
	case 0x03:
		return parse_03_msg(msg->buffer, msg->tail, msg->size);
#endif
#if CONFIG_MB_SLAVE_INPUT_REGISTERS
	case 0x04:
		return parse_03_msg(msg->buffer, msg->tail, msg->size);
#endif
#if CONFIG_MB_SLAVE_HOLDING_REGISTERS
	case 0x06:
		return parse_06_msg(msg->buffer, msg->tail);
	case 0x10:
		return parse_10_msg(msg->buffer, msg->tail);
#endif
	default:
		return send_error_message(msg->buffer, MBEX_NOSYS);
	}
}

void mb_slave_recv(void) __reentrant
{
	net_softirq();
}
